/*------------------------------------------------------------------------------------------
 																						
  Project:			Alpha Blending

  Module:			main.cpp - The windows overhead of the application.

  Author:			Matthias Hofmann											
																						
--------------------------------------------------------------------------------------------*/


#include "master.h"

//
// Globals.
//
LPDIRECTDRAWSURFACE7	osLogo = NULL;
LPDIRECTDRAWSURFACE7	osBackground = NULL;
LPDIRECTDRAWSURFACE7	osWork = NULL;
DWORD					g_dwBltType = 0;
DWORD					g_dwMode;
RECT					g_rcLogo;
bool					g_gActive;
bool					g_gIsMMX;
HFONT					g_hFont;


/*
 * Description:		The entry point for the application.
 *
 * Parameters:		As usual for windows applications.
 *
 * Return value:	The functions returns 0 to indicate 
 *					success or -1 if the call fails.
 *
 */
int WINAPI WinMain ( HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow )
{ 
	char			szTitle[]   = "Alpha Blending";	
	char			szAppName[] = "Alpha Blending"; 
    HWND			hWnd;
	MSG				msg;
	WNDCLASS		wc;


	//
	// Initialize and register window class
	//
    wc.style         = 0;	 
    wc.lpfnWndProc   = WindowProc;				 
    wc.cbClsExtra    = 0;						 
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance ;
    wc.hIcon         = LoadIcon( hInstance, MAKEINTRESOURCE ( IDI_MAIN ) ) ;
    wc.hCursor       = NULL;
    wc.hbrBackground = ( HBRUSH ) GetStockObject( NULL_BRUSH );
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = szAppName ;
    
    RegisterClass( &wc );

	//
	// Create the window for Space Blast
	//
	hWnd = CreateWindowEx(
        WS_EX_TOPMOST,					 
        szAppName,
        szTitle,
        WS_POPUP,						
        0, 0,							 
        GetSystemMetrics( SM_CXSCREEN ), 
        GetSystemMetrics( SM_CYSCREEN ), 
        NULL,
        NULL,
        hInstance,
        NULL );
	

	// Initialize the DirectDraw surfaces.
	if ( DDInit( hWnd ) == -1 )
		return -1;

	// Check for MMX capability.
	g_gIsMMX = ( GetCPUCaps() & CPU_FEATURE_MMX ) != 0;


	//
	// Show and immediately update window.
	//
    ShowWindow( hWnd, iCmdShow );
    UpdateWindow( hWnd );
	

	//
	// Message loop.
	//
	while ( true )
	{
		if ( PeekMessage( &msg, NULL, NULL, NULL, PM_REMOVE ) )
		{
			if ( msg.message == WM_QUIT )
			{ 
				// At this point, CleanUp() has already 
				// been called from the WindowProc.
				return msg.wParam;
			}

			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		else if ( !g_gActive )
		{
			WaitMessage();
		}
		else
		{
			UpdateFrame();
		}
	}
}

/*
 * Description:		Window procedure for the applications main window.
 *
 * Parameters:		As usual for window procs.
 *
 * Return value:	Depends on the message processed.
 *
 */
LRESULT CALLBACK WindowProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch ( msg )
    {
	case WM_CREATE:
		// Create a font.
		g_hFont = CreateFont( 28, 0, 0, 0, 600, false, false, false, 
			ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, 
			VARIABLE_PITCH | FF_DONTCARE, "Times New Roman" );
		break;

    case WM_ACTIVATEAPP:
		if ( wParam )
		{
			// The application is being activated.
			g_gActive = true;

			// If DirectDraw has already been initialized ...
			if ( lpDD7 )
			{
				// ... restore it.
				DDRestore();
			}
		}
		else
		{
			// The application is being deactivated.
			g_gActive = false;
		}
        return 0;

    case WM_KEYDOWN:
		//
		// Assign the blit type on keyboard input..
		//
        switch ( wParam )
        {
			case VK_F1:
				g_dwBltType = 0;
				return 0;

			case VK_F2:
				g_dwBltType = 1;
				return 0;

			case VK_F3:
				g_dwBltType = 2;
				return 0;

			case VK_F4:
				g_dwBltType = 3;
				return 0;

			case VK_F5:
				g_dwBltType = 4;
				return 0;

            case VK_ESCAPE:
				PostMessage( hWnd, WM_CLOSE, 0, 0 );
                return 0;
        }
        return 0;

	case WM_SETCURSOR:
		SetCursor( NULL );
		break;

	case WM_DESTROY:
		// Release the font.
		DeleteObject( g_hFont );

		// Clean up DirectDraw.
		DDCleanUp();

		// Signal to quit.
        PostQuitMessage( 0 );
        return 0;
    }

	return DefWindowProc( hWnd, msg, wParam, lParam );
}

/*
 * Description:		Updates the screen.
 *
 * Parameters:		None.
 *
 * Return value:	At the moment the function always returns 0.
 *
 */
int UpdateFrame( void )
{
	HDC				hdc;
	char			szBuffer[256];
	char*			szUnsupported = NULL;
	static float	iAlpha = 0;
	const int		iStep = 2;
	static int		iInc = iStep;
	static DWORD	dwTime;


	// Display the background.
	osWork->BltFast( 0, 0, osBackground, NULL, DDBLTFAST_WAIT );

	switch( g_dwBltType )
	{
	case 0:
		// Use the variable alpha function without MMX.
		BltAlpha( osWork, osLogo, 0, 0, &g_rcLogo, ( int ) iAlpha, g_dwMode );

		// Create the information string.
		wsprintf( szBuffer, "Using: BltAlpha()" );
		break;

	case 1:
		// If we got MMX support ...
		if ( g_gIsMMX )
		{
			// ... use the variable alpha function with MMX.
			BltAlphaMMX( osWork, osLogo, 0, 0, NULL, ( int ) iAlpha, g_dwMode );
		}
		// Otherwise ...
		else
		{
			// ... report that theres no MMX support.
			szUnsupported = "No MMX support!";
		}

		// Create the information string.
		wsprintf( szBuffer, "Using: BltAlphaMMX()" );
		break;

	case 2:
		// Use the fixed alpha function without MMX.
		BltAlphaFast( osWork, osLogo, 0, 0, NULL, g_dwMode );

		// Create the information string.
		wsprintf( szBuffer, "Using: BltAlphaFast()" );
		break;

	case 3:		   
		// If we got MMX support ...
		if ( g_gIsMMX )	
		{
			// ... use the fixed alpha function with MMX.
			BltAlphaFastMMX( osWork, osLogo, 0, 0, NULL, g_dwMode );
		}
		// Otherwise ...
		else
		{
			// ... report that theres no MMX support.
			szUnsupported = "No MMX support!";
		}

		// Create the information string.
		wsprintf( szBuffer, "Using: BltAlphaFastMMX()" );
		break;

	case 4:
		// Use the DirectDraw standard blit with color key.
		osWork->BltFast( 0, 0, osLogo, &g_rcLogo, DDBLTFAST_SRCCOLORKEY );

		// Create the information string.
		wsprintf( szBuffer, "Using: DirectX BltFast()" );
		break;
	}

	// Display the frame rate.
	ShowFrameRate();

	//
	// Prepare for GDI output.
	//
	osWork->GetDC( &hdc );
	SetBkMode( hdc, TRANSPARENT );
	SetTextColor( hdc, RGB( 0, 0, 128 ) );
	SelectObject( hdc, g_hFont );
	SetTextAlign( hdc, TA_TOP | TA_LEFT );

	// Display the current functions name string ...
	TextOut(hdc, 10, 10, szBuffer, strlen( szBuffer ) );

	SetTextAlign( hdc, TA_TOP | TA_CENTER );

	// Display the information string. 
	TextOut(hdc, 400, 550, "Press F1 to F5 to toggle the blitting functions.", 48 );

	//
	// Display an MMX message if necessary.
	//
	if ( szUnsupported )
	{
		TextOut(hdc, 400, 300, szUnsupported, strlen( szUnsupported ) );
	}

	// Finish GDI output.
	osWork->ReleaseDC(hdc);

	// Copy the working surface to the backbuffer.
	Back->BltFast( 0, 0, osWork, NULL, DDBLTFAST_WAIT );

	// Graphics done.
	Primary->Flip( NULL, DDFLIP_WAIT );

	
	// Increment the alpha value.
	iAlpha += ( signed ) ( iInc * ( timeGetTime() - dwTime ) ) * 60 / 1000.0f;

	// Update the timer.
	dwTime = timeGetTime();
	
	// If alpha is too high ...
	if ( iAlpha > 256 )
	{
		// ... limit it to its maximum ...
		iAlpha = 256;

		// ... and change the steps sign.
		iInc = -iStep;
	}
	// If alpha is too low ...
	else if ( iAlpha < 0 )
	{
		// ... limit it to its minimum ...
		iAlpha = 0;

		// ... and change the steps sign.
		iInc = iStep;
	}

	return 0;
}

/*
 * Description:		Initializes DirectDraw for the application.
 *
 * Parameters:		hWnd - The handle to the application window.
 *
 * Return value:	The functions returns 0 to indicate success 
 *					or -1 if the call fails.
 *
 */
int DDInit( HWND hWnd )
{
	SETTINGSINFO	si;
	SIZE			size;


	// Query the settings from the using.
	if ( QuerySettings( hWnd, &si ) == -1 )
		return -1;

	// Create the flipper.
	CreateFlipper( hWnd, &lpDD7, &si );

	//
	// Create a working surface.
	//
	size.cx = 800;
	size.cy = 600;
	CreateDDSurface( &osWork, &size, true );

	// Initialize the surface for the background.
	InitDDSurface( &osBackground, "background", true );

	// Initialize the surface for the logo.
	InitDDSurface( &osLogo, "logo", true ); 

	// Retrieve the size of the logo.
	GetPictureSize( "logo", &size );

	//
	// Initialize the logos rectangle.
	//
	g_rcLogo.left = 0;
	g_rcLogo.top = 0;
	g_rcLogo.right = size.cx;
	g_rcLogo.bottom = size.cy;

	// Get the exact display mode.
	g_dwMode = GetRGBMode( Primary );

	return 0;
}

/*
 * Description:		Restores DirectDraw for the application.
 *
 * Parameters:		None.
 *
 * Return value:	None.
 *
 */
void DDRestore( void )
{
	//
	// Restore the flipper.
	//
	if ( Primary )
	{
		if ( Primary->IsLost() == DDERR_SURFACELOST )
			Primary->Restore();	// This restores the backbuffer as well.
	}

	//
	// Restore the working surface.
	//
	if ( osWork->IsLost() == DDERR_SURFACELOST )
		osWork->Restore();

	// Restore the surface for the background.
	InitDDSurface( &osBackground, "background", true );

	// Restore the surface for the logo.
	InitDDSurface( &osLogo, "logo", true ); 
}

/*
 * Description:		Cleans up DirectDraw.
 *
 * Parameters:		None.
 *
 * Return value:	None.
 *
 */
void DDCleanUp( void )
{
	//
	// Release the working surface.
	//
	if ( osWork )
	{
		osWork->Release();
		osWork = NULL;
	}

	//
	// Release the surface for the background.
	//
	if ( osBackground )
	{
		osBackground->Release();
		osBackground = NULL;
	}

	//
	// Release the surface for the logo.
	//
	if ( osLogo )
	{
		osLogo->Release();
		osLogo = NULL;
	}

	// 
	// Release the flipper.
	//
	if ( Primary )
	{
		Primary->Release();
		Primary = NULL;
	}
}


